#ifndef ASSOCIATIVE_ARRAY_INCLUDED
#define ASSOCIATIVE_ARRAY_INCLUDED

const WInt KeyValueAlreadyExists = -1;
const WInt KeyValueNotFound      = -2;
const WInt ExecutingForAll       = -3;

template<class KeyType,class ElementType>
class AssociativeArray
{
    private: 
             struct ArrayElement
             {
                 ElementType element_value;
                 KeyType     key_value;
             };
    
    private: 
             WArray<ArrayElement> table;
             WUInt                execution_depth;
    
    private: 
             WUInt GetKeyPosition (const KeyType key_value) const;
    
    public : 
             AssociativeArray ();
    
    public : 
             const ElementType& GetAtPosition (const WUInt position)    const;
             const ElementType& operator []   (const KeyType key_value) const;
    
    public : 
             WUInt GetNumberElements      ()                        const;
             WBool KeyValueDefined (const KeyType key_value) const;
    
    public : 
             void AddElement (const KeyType key_value,const ElementType element_value);
    
    public : 
             typedef WBool WCMDEF (*ActionFunction) (const KeyType&,ElementType&);
    
    public : 
             WBool ExecuteForAll (ActionFunction action);
};

template<class KeyType,class ElementType>
void AssociativeArray<KeyType,ElementType>::AddElement (const KeyType key_value,const ElementType element_value)
{
    if (execution_depth > 0)
        throw ExecutingForAll;
        
    const WUInt number_elements = table.GetCount ();
    const WUInt position        = GetKeyPosition (key_value);
    
    if (position < number_elements && table [position].key_value == key_value)
        throw KeyValueAlreadyExists;
    
    table.SetCount (number_elements + 1);
    
    for (WUInt index = number_elements; index > position; index--)
        table [index] = table [index - 1];
    
    table [position].element_value = element_value;
    table [position].key_value     = key_value;
}

template<class KeyType,class ElementType>
AssociativeArray<KeyType,ElementType>::AssociativeArray ()
{
    execution_depth = 0;
}

template<class KeyType,class ElementType>
WBool AssociativeArray<KeyType,ElementType>::ExecuteForAll (ActionFunction action)
{
    execution_depth++;
    
    for (WInt index = table.GetCount () - 1; index >= 0; index--)
        if (!action (table [index].key_value,table [index].element_value))
        {
            execution_depth--;
            return (FALSE);
        }
    
    execution_depth--;
    return (TRUE);
}

template<class KeyType,class ElementType>
const ElementType& AssociativeArray<KeyType,ElementType>::GetAtPosition (const WUInt position)    const
{
    WASSERTEX (position < table.GetCount (),"Invalid associative array position.");
    return    (table [position].element_value);
}

template<class KeyType,class ElementType>
WUInt AssociativeArray<KeyType,ElementType>::GetKeyPosition (const KeyType key_value) const
{
    const WUInt number_elements = table.GetCount ();
    
    for (int index = 0; index < number_elements; index++)
        if (table [index].key_value >= key_value)
            return (index);
    
    return (number_elements);
}

template<class KeyType,class ElementType>
WUInt AssociativeArray<KeyType,ElementType>::GetNumberElements () const
{
    return (table.GetCount ());
}

template<class KeyType,class ElementType>
WBool AssociativeArray<KeyType,ElementType>::KeyValueDefined (const KeyType key_value) const
{
    const WUInt position = GetKeyPosition (key_value);
    return (position < table.GetCount () && table [position].key_value == key_value);
}

template<class KeyType,class ElementType>
const ElementType& AssociativeArray<KeyType,ElementType>::operator [] (const KeyType key_value) const
{
    const WUInt position = GetKeyPosition (key_value);
    
    if (position < table.GetCount () && table [position].key_value == key_value)
        return (table [position].element_value);
    else
        throw KeyValueNotFound;
}

#endif

