#ifndef LEXICAL_INCLUDED
#define LEXICAL_INCLUDED

#include "FileError.hpp"
#include "File.hpp"

const WInt  InvalidDFAStateNumber = -1;
const WInt  InvalidTokenValue     = -1;
const WInt  EndOfFileToken        =  0;
const WChar EndOfFileSymbol       = (WChar) EOF;
/*
    DESCRIP: InvalidDFAStateNumber -- Used internally to represent a state which does
                                      not exist.
             InvalidTokenValue     -- Value returned when the simulator cannot recognize
                                      anything from the input.
             EndOfFileToken        -- Value returned when the simulator recognizes the
                                      EndOfFileSymbol.
             EndOfFileSymbol       -- Value which should be returned by the derived class
                                      when the end of the input has been reached.
*/

/*
    DFATransitionStruct -- Internal representation of a transition from a DFA state.
    DFAStateStruct      -- Internal representation of a DFA state.
    LexicalBase         -- Pure abstract class which forms the base of a lexical analyzer.
                           Derived class must supply functions get read the input stream
                           and initialize the automaton data structures.
*/

#pragma pack (push,1) // Ensure single byte alignment for subsequent declarations

/*
    The Optima++ library defines the following type.  When compiling under another
    system, please alter the following definitions appropriately and have your compiler
    define the OPTIMA_TYPES_NOT_DEFINED symbol before compiling any files which require
    this file.  Note that if you're using the LexicalBase.HPP file, it's better to modify
    those typedefs and ignore these.
*/
#ifdef    OPTIMA_TYPES_NOT_DEFINED
#undefine OPTIMA_TYPES_NOT_DEFINED
      typedef unsigned int  WUInt;        // 32-bit unsigned integer
      typedef          char WChar;        //  8-bit character
      typedef          int  WInt;         // 32-bit signed integer
//    typedef ????????      WString;      // Variable length string which supports
                                          // the += operator to add a character to the
                                          // end of the string.  Most C++ libraries
                                        // will have something similar, but, if you
                                        // change it to a fixed length array of char,
                                        // be sure to modify the GetNextToken function
                                        // so that it cannot over-run the array boundary.
#endif

struct DFATransitionStruct
{
    WChar symbol;
    WUInt state;
};

struct DFAStateStruct
{
    WInt first_transition,number_transitions;
    WInt accepted_token;
};
#pragma pack (pop)    // Restore original alignment

class LexicalBase
{
    protected: 
               virtual WChar GetCharacterLookahead () = 0;
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- Routine should non-destructively return the next
                                       character from the input stream.
               */
               
               virtual WChar GetNextCharacter      () = 0;
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- Routine should remove and return the next character
                                       from the input stream.
               */
    
    protected: 
                DFATransitionStruct* transition_table;
                DFAStateStruct*      state_table;
               /*
                   DESCRIP: transition_table -- All transitions in the automaton.
                            state_table      -- All states in the automaton.
                   NOTE   : These arrays must be allocated, with the 'new' operator,
                            and initialized by the derived class.
               */
    
    private  : 
               WInt GetTransitionFromState (const WUInt current_state,const WChar symbol) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- The state to which a transition occurs on
                                       the given symbol from the given state, or
                                       InvalidDFAStateNumber if no such transition
                                       is possible.
               */
               {
                   for (WInt index = state_table [current_state].first_transition + state_table [current_state].number_transitions - 1; index >= state_table [current_state].first_transition; index--)
                       if (transition_table [index].symbol == symbol)
                           return (transition_table [index].state);
                   
                   return (InvalidDFAStateNumber);
               }
    
    private  : 
                WString lookahead_lexeme;
                WBool   lookahead_valid;
                WInt    lookahead_token;
               /*
                   DESCRIP: Look-ahead information.  Only valid if 'lookahead_valid'
                            is TRUE.
               */
    
    public   : 
               LexicalBase ()
               {
                   transition_table = NULL;
                   lookahead_valid  = FALSE;
                   state_table      = NULL;
               }
               
               virtual ~LexicalBase ()
               {
                   delete[] transition_table;
                   delete[] state_table;
               }
    
    protected: 
               WBool IsValidDFAState (const DFAStateStruct& state,const WUInt number_transitions) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- TRUE if the given state is valid given the number
                                       of transitions in the automaton.
               */
               {
                   return (state.first_transition >= 0 && state.first_transition + state.number_transitions <= number_transitions);
               }
               
               WBool IsValidDFATransition (const DFATransitionStruct& transition,const WUInt number_states) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- TRUE if the given transition is valid given the 
                                       number of states in the automaton.
               */
               {
                   return (transition.state < number_states && ((transition.symbol >= ' ' && transition.symbol <= '~') || (transition.symbol == '\f' || transition.symbol == '\n' || transition.symbol == '\r' || transition.symbol == '\t' || transition.symbol == EndOfFileSymbol)));
               }
    
    
    public   : 
               WInt GetNextToken (WString *const lexeme)
               /*
                   PRECOND : lexeme   -- Pointer to a string which will contain the lexeme
                                         which is read.  NULL if the lexeme is not needed.
                   POSTCOND: INSTANCE -- Either the look-ahead information or characters
                                         from the input stream are consumed.
                             RETURN   -- The value of the token recognized.
                             lexeme   -- If not NULL, contains the lexeme read.
               */
               {
                   WString token_lexeme;
                   WInt    token_value;
                   
                   if (!lookahead_valid)
                   {
                       while (isspace (GetCharacterLookahead ()))
                           GetNextCharacter ();
                       
                       WUInt current_state = 0;
                       WInt  next_state    = GetTransitionFromState (current_state,GetCharacterLookahead ());
                       
                       while (next_state != InvalidDFAStateNumber)
                       {
                           token_lexeme += GetNextCharacter ();
                           current_state = next_state;
                           next_state    = GetTransitionFromState (current_state,GetCharacterLookahead ());
                       }
                       
                       if (token_lexeme == "")
                           token_lexeme += GetNextCharacter ();
                       
                       token_value = state_table [current_state].accepted_token;
                   }
                   else
                   {
                       lookahead_valid = FALSE;
                       token_lexeme    = lookahead_lexeme;
                       token_value     = lookahead_token;
                   }
                   
                   if (lexeme != NULL)
                       *lexeme = token_lexeme;
                   
                   return (token_value);
               }
               
               WInt GetTokenLookahead ()
               /*
                   PRECOND : 
                   POSTCOND: INSTANCE -- Look-ahead information is valid.
                             RETURN   -- Value of the next token in the input stream.
               */
               {
                   lookahead_token = GetNextToken (&lookahead_lexeme);
                   lookahead_valid = TRUE;
                   
                   return (lookahead_token);
               }
};


#endif

