#ifndef PARSER_INCLUDED
#define PARSER_INCLUDED

const WInt InvalidSLRStateNumber = -1;

enum ActionTableMnemonic {AcceptAction,ErrorAction,ReduceAction,ShiftAction};

#pragma pack (push,1)    // Single byte alignment

/*
    The Optima++ library defines the following type.  When compiling under another
    system, please alter the following definitions appropriately and have your compiler
    define the OPTIMA_TYPES_NOT_DEFINED symbol before compiling any files which require
    this file.  Note that if you're using the LexicalBase.HPP file, it's better to modify
    those typedefs and ignore these.
*/
#ifdef    OPTIMA_TYPES_NOT_DEFINED
#undefine OPTIMA_TYPES_NOT_DEFINED
      typedef unsigned int  WUInt;        // 32-bit unsigned integer
      typedef          char WChar;        //  8-bit character
      typedef          int  WInt;         // 32-bit signed integer
//    typedef ????????      WString;      // See LexicalBase.HPP
#endif

/*
    SLRActionStruct -- Represents an entry in the sparse action table with 'terminal'
                       determining the index of the entry.  If 'action' is ShiftAction,
                       'general' is the number of the state to shift.  If it is 
                       ReduceAction, 'general' is the number of the non-terminal to
                       reduce and 'length' is the length of the right hand side of the
                       production being reduced.
*/
struct SLRActionStruct
{
    ActionTableMnemonic action;
    WUInt               terminal;
    WUInt               general;
    WUInt               length;
};

struct SLRGotoStruct
{
    WUInt non_terminal;
    WUInt state_number;
};

struct SLRStateStruct
{
    WUInt number_actions,number_gotos;
    WUInt first_action,first_goto;
};
#pragma pack (pop)    // Restore original alignment

/*
    ParseTreeNode -- Most basic information about a node in the parse tree.  Only
                     stores whether it's a terminal or non-terminal, the symbol
                     number, and its right sibling in the tree.  Derive more detailed
                     types directly or from the TerminalNode and NonTerminalNode classes,
                     below.
*/
class ParseTreeNode
{
    private: 
                   ParseTreeNode* right_sibling;
             const WUInt          symbol_number;
             const WBool          non_terminal;
    
    public : 
             ParseTreeNode (const WUInt symbol,const WBool terminal) : non_terminal (!terminal),symbol_number (symbol)
             {
                 right_sibling = NULL;
             }
             
             virtual ~ParseTreeNode ()
             {
                 delete right_sibling;
             }
    
    public : 
             ParseTreeNode *const GetSibling ()
             {
                 return (right_sibling);
             }
             
             WUInt GetSymbolNumber () const
             {
                 return (symbol_number);
             }
             
             WBool IsNonTerminal () const
             {
                 return (non_terminal);
             }
             
             void SetRightSibling (ParseTreeNode* sibling)
             {
                 WASSERT (right_sibling == NULL);
                 right_sibling = sibling;
             }
};

/*
    TerminalNode -- Very simple representation of a terminal symbol in the parse tree.
                    Stores the symbol number, symbol lexeme, and right sibling.
*/
class TerminalNode : public ParseTreeNode
{
    private: 
             const WString symbol_lexeme;
    
    public : 
             TerminalNode (const WUInt symbol_number,const WString lexeme) : ParseTreeNode (symbol_number,TRUE),symbol_lexeme (lexeme)
             {
             }
    
    public : 
             WString GetTerminalLexeme () const
             {
                 return (symbol_lexeme);
             }
};

/*
    NonTerminalNode -- Very simple representation of a non-terminal symbol in the parse
                       tree.  Stores symbol number, right sibling, and children of
                       the non-terminal (i.e., its production).
*/
class NonTerminalNode : public ParseTreeNode
{
    private: 
             ParseTreeNode* leftmost_child;
    
    public : 
             NonTerminalNode (const WUInt symbol_number,ParseTreeNode *const production) : ParseTreeNode (symbol_number,FALSE),leftmost_child (production)
             {
             }
             
             virtual ~NonTerminalNode ()
             {
                 delete leftmost_child;
             }
    
    public : 
             ParseTreeNode *const GetProduction ()
             {
                 return (leftmost_child);
             }
};

/*
    ParsingStack -- Stack used during processing of an LR parsing table.
*/
class ParsingStack
{
    private: 
             ParseTreeNode** symbol_stack;
             WUInt*          state_stack;
             WUInt           stack_size,stack_top;
             /*
                 DESCRIP: symbol_stack -- Grammar symbols currently on the stack.
                                          Because there is always one more state than
                                          grammar symbol, the top element of this stack
                                          is two less than 'stack_top'.
                          state_stack  -- States currently on the stack.  Top element
                                          is one less than 'stack_top'.
                          stack_size   -- Maximum number of states which may be pushed
                                          onto the stack.
                          stack_top    -- Number of states currently on the stack.
             */
    
    public : 
             ParsingStack ()
             /*
                 PRECOND : 
                 POSTCOND: INSTANCE -- Creates a stack with state number zero
                                       and no symbols.
             */
             {
                 const WUInt InitialStackSize = 32;
                 
                 symbol_stack = new ParseTreeNode* [InitialStackSize];
                 state_stack  = new WUInt          [InitialStackSize];
                 stack_size   = InitialStackSize;
                 stack_top    = 1;
                 
                 state_stack [0] = 0;
             }
             
             ~ParsingStack ()
             /*
                 PRECOND : 
                 POSTCOND: INSTANCE -- If parsing ended prematurly, any symbols
                                       still on the stack are deleted.  In any case,
                                       both stacks are deleted.
             */
             {
                 for (WInt index = stack_top - 2; index >= 0; index--)
                     delete symbol_stack [index];
                 
                 delete[] symbol_stack;
                 delete[] state_stack;
             }
    
    public : 
             WUInt GetStateStackTop () const
             /*
                 PRECOND : 
                 POSTCOND: RETURN -- The state number at the top of the stack.
             */
             {
                 return (state_stack [stack_top - 1]);
             }
             
             ParseTreeNode* GetSymbolStackTop ()
             /*
                 PRECOND : INSTANCE -- Nothing should have been added to the stack since
                                       the last call to RemoveProduction.
                 POSTCOND: RETURN   -- A pointer to the symbol (as a ParseTreeNode) at
                                       the top of the stack.
             */
             {
                 return (symbol_stack [stack_top - 2]);
             }
             
             ParseTreeNode* RemoveProduction (const WUInt production_length)
             /*
                 PRECOND : production_length -- The number of symbols and states to
                                                remove from the stack.
                 POSTCOND: RETURN            -- A pointer to the first element in
                                                the production.
                           INSTANCE          -- 'production_length' symbols are removed
                                                from the symbol stack and from the state
                                                stack.
             */
             {
                 if (production_length == 0)
                     return (NULL);
                     
                 stack_top -= production_length;
                 
                 for (WUInt index = 0; index < production_length; index++)
                     symbol_stack [stack_top - 1 + index]->SetRightSibling (index + 1 < production_length ? symbol_stack [stack_top + index] : NULL);
                 
                 return (symbol_stack [stack_top - 1]);
             }
    
    public : 
             void PushSymbol (ParseTreeNode* symbol,const WUInt state)
             /*
                 PRECOND : 
                 POSTCOND: INSTANCE -- Pushes the given symbol and its covering state
                                       onto the stack.
             */
             {
                 symbol_stack [stack_top - 1] = symbol;
                 
                 if (stack_top == stack_size)
                 {
                     const WUInt StackSizeIncrease = 16;
                     
                     ParseTreeNode** temp_symbols = new ParseTreeNode* [stack_size + StackSizeIncrease];
                     WUInt*          temp_states  = new WUInt          [stack_size + StackSizeIncrease];
                     
                     memcpy (temp_symbols,symbol_stack,stack_size * sizeof (ParseTreeNode*));
                     memcpy (temp_states ,state_stack ,stack_size * sizeof (WUInt));
                     
                     delete[] symbol_stack;
                     delete[] state_stack;
                     
                     symbol_stack = temp_symbols;
                     state_stack  = temp_states;
                 }
                 
                 state_stack [stack_top] = state;
                 stack_top              += 1;
             }
};

/*
    ParserBase -- Abstract class for executing an LR parsing table.  Derived class must
                  initialize 'action_table', 'state_table' and 'goto_table' and provide
                  functions to get the next token in the input stream, reduce terminals
                  and non-terminals, and take error correction actions.
*/
class ParserBase
{
    protected: 
               SLRActionStruct* action_table;
               SLRStateStruct*  state_table;
               SLRGotoStruct*   goto_table;
               /*
                   DESCRIP: action_table -- Sparse table of SLRActionStructs.
                            state_table  -- Complete table of SLRStateStructs.
                            goto_table   -- Sparse table of SLRGotoStructs.
                   NOTE   : These three tables must be initialized by the derived
                            class.
               */
    
    protected: 
               virtual WInt GetNextToken      (WString& lexeme) = 0;
               /*
                   PRECOND : 
                   POSTCOND: lexeme - The lexeme of the token returned.
                             RETURN - Removes the next token from the input stream
                                      and returns its value.
               */
               
               virtual WInt GetTokenLookahead ()                = 0;
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- Returns next the value of the next token in
                                       the input stream without removing it.
               */
    
    protected: 
               virtual ParseTreeNode* RecognizeNonTerminal (const WUInt symbol_number,ParseTreeNode *const production) = 0;
               /*
                   PRECOND : symbol_number -- Number of the non-terminal being reduced.
                             production    -- Right-hand side of the production being
                                              reduced.
                   POSTCOND: RETURN        -- A pointer to a ParseTreeNode (or derived
                                              class) storing information about the 
                                              non-terminal, or NULL to cancel parsing.
               */
               
               virtual ParseTreeNode* RecognizeTerminal (const WUInt symbol_number,const WString symbol_lexeme) = 0;
               /*
                   PRECOND : symbol_number -- Number of the terminal symbol being
                                              recognized.
                             lexeme        -- Lexeme of the terminal symbol.
                   POSTCOND: RETURN        -- A pointer to a ParseTreeNode (or derived
                                              class) storing information about the
                                              terminal, or NULL to cancel parsig.
               */
               
               virtual WBool ErrorCorrected (ParsingStack& parsing_stack,const WInt token_lookahead) = 0;
               /*
                   PRECOND : parsing_stack   -- The parsing stack as it was when the error
                                                was recognized.
                             token_lookahead -- The token for which there was no action
                                                record given the state on top of the stack.
                   POSTCOND: parsing_stack   -- May have been modified in accordance with
                                                an error correction scheme.
                             INSTANCE        -- May modify the input stream in accordance
                                                with an error correction scheme.
                             RETURN          -- TRUE if parsing should continue.  FALSE
                                                to cancel.
               */
    
    protected:
               WBool IsValidSLRAction (const SLRActionStruct& action,const WUInt number_states) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- TRUE if 'action' appears to be valid given the 
                                       number of states in the table.
               */
               {
                   switch (action.action)
                   {
                       case AcceptAction: return (TRUE);
                       case ReduceAction: return (TRUE);
                       case ShiftAction : return (action.general < number_states);
                       default          : return (FALSE);
                   }
               }
               
               WBool IsValidSLRGoto (const SLRGotoStruct& goto_value,const WUInt number_states) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- TRUE if 'goto_value' appears to be valid given
                                       the number of states in the table.
               */
               {
                   return (goto_value.state_number < number_states);
               }
               
               WBool IsValidSLRState (const SLRStateStruct& state,const WUInt number_actions,const WUInt number_gotos) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- TRUE if 'state' appears to be valid given the
                                       number of actions and gotos in the table.
               */
               {
                   return (state.first_action + state.number_actions <= number_actions && state.first_goto + state.number_gotos <= number_gotos);
               }
    
    private  : 
               SLRActionStruct GetActionTableEntry (const WUInt state,const WUInt terminal_symbol) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- The action to take given the state number and
                                       terminal symbol.  ErrorAction if no other actions
                                       are found.
               */
               {
                   const static SLRActionStruct  error_action = {ErrorAction,0xFFFF,0xFFFF,0xFFFF};
                                SLRActionStruct* action_ptr   = &action_table [state_table [state].first_action];
                   
                   for (WUInt index = 0; index < state_table [state].number_actions; index++,action_ptr++)
                       if (action_ptr->terminal == terminal_symbol)
                           return (*action_ptr);
                   
                   return (error_action);
               }
               
               WUInt GetGotoState (const WUInt state_number,const WUInt non_terminal) const
               /*
                   PRECOND : 
                   POSTCOND: RETURN -- Give the state, returns the state number which 
                                       should be pushed on top of the non-terminal.
               */
               {
                   SLRGotoStruct* goto_ptr = &goto_table [state_table [state_number].first_goto];
                   
                   for (WUInt index = 0; index < state_table [state_number].number_gotos; index++,goto_ptr++)
                       if (goto_ptr->non_terminal == non_terminal)
                           return (goto_ptr->state_number);
                   
                   return (0xFFFF);
               }
    
    public   : 
               ParserBase ()
               {
                   action_table = NULL;
                   state_table  = NULL;
                   goto_table   = NULL;
               }
               
               virtual ~ParserBase ()
               {
                   delete[] action_table;
                   delete[] state_table;
                   delete[] goto_table;
               }
    
    public   : 
               ParseTreeNode* ExecuteParser ()
               /*
                   PRECOND : INSTANCE -- Derived type has initialized the appropriate
                                         tables.
                   POSTCOND: RETURN   -- A pointer to the root node of a parse tree
                                         if parsing is successful, or NULL if an
                                         uncorrected error occurs.
               */
               {
                   ParsingStack parsing_stack;
                   
                   do
                   {
                       const SLRActionStruct entry = GetActionTableEntry (parsing_stack.GetStateStackTop (),GetTokenLookahead ());
                       
                       switch (entry.action)
                       {
                           case AcceptAction:
                           {
                               ParseTreeNode* symbol = parsing_stack.GetSymbolStackTop ();
                               
                               parsing_stack.RemoveProduction (1);
                               return                         (symbol);
                           }
                           
                           case ReduceAction: 
                           {
                               ParseTreeNode* symbol = RecognizeNonTerminal (entry.general,parsing_stack.RemoveProduction (entry.length));
                               
                               if (symbol != NULL)
                                   parsing_stack.PushSymbol (symbol,GetGotoState (parsing_stack.GetStateStackTop (),entry.general));
                               else
                                   return (NULL);
                               
                               break;
                           }
                           
                           case ShiftAction : 
                           {
                               WString        token_lexeme;
                               WInt           token_number = GetNextToken (token_lexeme);
                               ParseTreeNode* symbol       = RecognizeTerminal (token_number,token_lexeme);
                               
                               if (symbol != NULL)
                                   parsing_stack.PushSymbol (symbol,entry.general);
                               else
                                   return (NULL);
                                   
                               break;
                           }
                           
                           case ErrorAction : 
                           {
                               if (!ErrorCorrected (parsing_stack,GetTokenLookahead ()))
                                   return (NULL);
                               
                               break;
                           }
                           
                           default: 
                               WWARNING ("Invalid SLRActionStruct in ParserBase::ExecuteParser.");
                               return   (NULL);
                       }
                   }
                   while (TRUE);
               }
};

#endif

